import copy
import functools
import importlib.metadata
import importlib.util
import json
import logging
import os
import sys

import click
from google import auth

from octue.cloud import storage
from octue.cloud.pub_sub import Subscription, Topic
from octue.cloud.pub_sub.service import Service
from octue.cloud.service_id import convert_service_id_to_pub_sub_form, create_sruid, get_sruid_parts
from octue.cloud.storage import GoogleCloudStorageClient
from octue.configuration import load_service_and_app_configuration
from octue.definitions import MANIFEST_FILENAME, VALUES_FILENAME
from octue.exceptions import ServiceAlreadyExists
from octue.log_handlers import apply_log_handler, get_remote_handler
from octue.resources import Manifest, service_backends
from octue.runner import Runner
from octue.utils.encoders import OctueJSONEncoder
from twined import Twine


logger = logging.getLogger(__name__)

global_cli_context = {}


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "--id",
    default=None,
    type=click.UUID,
    show_default=True,
    help="UUID of the analysis being undertaken. None (for local use) will cause a unique ID to be generated.",
)
@click.option("--logger-uri", default=None, show_default=True, help="Stream logs to a websocket at the given URI.")
@click.option(
    "--log-level",
    default="info",
    type=click.Choice(["debug", "info", "warning", "error"], case_sensitive=False),
    show_default=True,
    help="Log level used for the analysis.",
)
@click.option(
    "--force-reset/--no-force-reset",
    default=True,
    is_flag=True,
    show_default=True,
    help="Forces a reset of analysis cache and outputs [For future use, currently not implemented]",
)
@click.version_option(version=importlib.metadata.version("octue"))
def octue_cli(id, logger_uri, log_level, force_reset):
    """The CLI for the Octue SDK. Use it to start an Octue data service or digital twin locally or run an analysis on
    one locally.

    Read more in the docs: https://octue-python-sdk.readthedocs.io/en/latest/
    """
    global_cli_context["analysis_id"] = id
    global_cli_context["logger_uri"] = logger_uri
    global_cli_context["log_handler"] = None
    global_cli_context["log_level"] = log_level.upper()
    global_cli_context["force_reset"] = force_reset

    apply_log_handler(log_level=log_level.upper())

    if global_cli_context["logger_uri"]:
        global_cli_context["log_handler"] = get_remote_handler(logger_uri=global_cli_context["logger_uri"])


@octue_cli.command()
@click.option(
    "-c",
    "--service-config",
    type=click.Path(dir_okay=False),
    default="octue.yaml",
    help="The path to an `octue.yaml` file defining the service to run.",
)
@click.option(
    "--input-dir",
    type=click.Path(file_okay=False, exists=True),
    default=".",
    show_default=True,
    help="The path to a directory containing the input values (in a file called 'values.json') and/or input manifest "
    "(in a file called 'manifest.json').",
)
@click.option(
    "-o",
    "--output-file",
    type=click.Path(dir_okay=False),
    default=None,
    show_default=True,
    help="The path to a JSON file to store the output values in, if required.",
)
@click.option(
    "--output-manifest-file",
    type=click.Path(dir_okay=False),
    default=None,
    help="The path to a JSON file to store the output manifest in. The default is 'output_manifest_<analysis_id>.json'.",
)
@click.option(
    "--monitor-messages-file",
    type=click.Path(dir_okay=False),
    default=None,
    show_default=True,
    help="The path to a JSON file in which to store any monitor messages received. Monitor messages will be ignored "
    "if this option isn't provided.",
)
def run(service_config, input_dir, output_file, output_manifest_file, monitor_messages_file):
    """Run an analysis on the given input data using an Octue service or digital twin locally. The output values are
    printed to `stdout`. If an output manifest is produced, it will be saved locally (see the `--output-manifest-file`
    option).
    """
    service_configuration, app_configuration = load_service_and_app_configuration(service_config)

    input_values_path = os.path.join(input_dir, VALUES_FILENAME)
    input_manifest_path = os.path.join(input_dir, MANIFEST_FILENAME)

    input_values = None
    input_manifest = None

    if os.path.exists(input_values_path):
        input_values = input_values_path

    if os.path.exists(input_manifest_path):
        input_manifest = input_manifest_path

    runner = Runner(
        app_src=service_configuration.app_source_path,
        twine=Twine(source=service_configuration.twine_path),
        configuration_values=app_configuration.configuration_values,
        configuration_manifest=app_configuration.configuration_manifest,
        children=app_configuration.children,
        output_location=app_configuration.output_location,
        crash_diagnostics_cloud_path=service_configuration.crash_diagnostics_cloud_path,
        service_registries=service_configuration.service_registries,
    )

    if monitor_messages_file:
        if not os.path.exists(os.path.dirname(monitor_messages_file)):
            os.makedirs(os.path.dirname(monitor_messages_file))

        monitor_message_handler = lambda message: _add_monitor_message_to_file(monitor_messages_file, message)

    else:
        monitor_message_handler = None

    analysis = runner.run(
        analysis_id=global_cli_context["analysis_id"],
        input_values=input_values,
        input_manifest=input_manifest,
        analysis_log_level=global_cli_context["log_level"],
        analysis_log_handler=global_cli_context["log_handler"],
        handle_monitor_message=monitor_message_handler,
    )

    click.echo(json.dumps(analysis.output_values, cls=OctueJSONEncoder))

    if analysis.output_values and output_file:
        if not os.path.exists(os.path.dirname(output_file)):
            os.makedirs(os.path.dirname(output_file))

        with open(output_file, "w") as f:
            json.dump(analysis.output_values, f, cls=OctueJSONEncoder, indent=4)

    if analysis.output_manifest:
        if not os.path.exists(os.path.dirname(output_manifest_file)):
            os.makedirs(os.path.dirname(output_manifest_file))

        with open(output_manifest_file or f"output_manifest_{analysis.id}.json", "w") as f:
            json.dump(analysis.output_manifest.to_primitive(), f, cls=OctueJSONEncoder, indent=4)

    return 0


@octue_cli.command()
@click.option(
    "-c",
    "--service-config",
    type=click.Path(dir_okay=False),
    default="octue.yaml",
    help="The path to an `octue.yaml` file defining the service to start.",
)
@click.option(
    "--revision-tag",
    type=str,
    default=None,
    help="A tag to use for this revision of the service (e.g. 1.3.7). This overrides the `OCTUE_SERVICE_REVISION_TAG` "
    "environment variable if it's present. If this option isn't given and the environment variable isn't present, a "
    "random 'cool name' tag is generated e.g 'curious-capybara'.",
)
@click.option(
    "--timeout",
    type=click.INT,
    default=None,
    show_default=True,
    help="A timeout in seconds after which to stop the service. The default is no timeout.",
)
@click.option(
    "--no-rm",
    is_flag=True,
    default=False,
    show_default=True,
    help="Don't delete the Google Pub/Sub topic and subscription for the service on exit.",
)
def start(service_config, revision_tag, timeout, no_rm):
    """Start an Octue service or digital twin locally as a child so it can be asked questions by other Octue services.
    The service's pub/sub topic and subscription are deleted on exit.
    """
    service_revision_tag_override = revision_tag
    service_configuration, app_configuration = load_service_and_app_configuration(service_config)
    service_namespace, service_name, service_revision_tag = get_sruid_parts(service_configuration)

    if service_revision_tag_override and service_revision_tag:
        logger.warning(
            "The `OCTUE_SERVICE_REVISION_TAG` environment variable %r has been overridden by the `--revision-tag` CLI "
            "option %r.",
            service_revision_tag,
            service_revision_tag_override,
        )

    service_sruid = create_sruid(
        namespace=service_namespace,
        name=service_name,
        revision_tag=service_revision_tag_override or service_revision_tag,
    )

    runner = Runner(
        app_src=service_configuration.app_source_path,
        twine=Twine(source=service_configuration.twine_path),
        configuration_values=app_configuration.configuration_values,
        configuration_manifest=app_configuration.configuration_manifest,
        children=app_configuration.children,
        output_location=app_configuration.output_location,
        crash_diagnostics_cloud_path=service_configuration.crash_diagnostics_cloud_path,
        service_id=service_sruid,
    )

    run_function = functools.partial(
        runner.run,
        analysis_log_level=global_cli_context["log_level"],
        analysis_log_handler=global_cli_context["log_handler"],
    )

    backend_configuration_values = (app_configuration.configuration_values or {}).get("backend")

    if backend_configuration_values:
        backend_configuration_values = copy.deepcopy(backend_configuration_values)
        backend = service_backends.get_backend(backend_configuration_values.pop("name"))(**backend_configuration_values)
    else:
        # If no backend details are provided, use Google Pub/Sub with the default project.
        _, project_name = auth.default()
        backend = service_backends.get_backend()(project_name=project_name)

    service = Service(service_id=service_sruid, backend=backend, run_function=run_function)

    try:
        service.serve(timeout=timeout, delete_topic_and_subscription_on_exit=not no_rm)

    except ServiceAlreadyExists:
        # Generate and use a new revision tag if the service already exists.
        service_sruid = create_sruid(namespace=service_namespace, name=service_name)

        while True:
            user_confirmation = input(
                "Service already exists. Create new service with service revision unique identifier (SRUID) "
                f"{service_sruid!r}? [Y/n]\n"
            )

            if user_confirmation.upper() == "N":
                return

            if user_confirmation.upper() in {"Y", ""}:
                break

        service = Service(service_id=service_sruid, backend=backend, run_function=run_function)
        service.serve(timeout=timeout, delete_topic_and_subscription_on_exit=not no_rm)


@octue_cli.command()
@click.argument(
    "cloud_path",
    type=str,
)
@click.option(
    "--local-path",
    type=click.Path(file_okay=False),
    default=".",
    help="The path to a directory to store the directory of diagnostics data in. Defaults to the current working "
    "directory.",
)
@click.option(
    "--download-datasets",
    is_flag=True,
    help="If provided, download any datasets from the crash diagnostics and update their paths in the configuration and "
    "input manifests to the new local paths.",
)
def get_crash_diagnostics(cloud_path, local_path, download_datasets):
    """Download crash diagnostics for an analysis from the given directory in Google Cloud Storage. The cloud path
    should end in the analysis ID.

    CLOUD_PATH: The path to the directory in Google Cloud Storage containing the diagnostics data.
    """
    analysis_id = storage.path.split(cloud_path)[-1]
    local_path = os.path.join(local_path, analysis_id)

    if download_datasets:
        filter = None
    else:
        filter = lambda blob: any(
            (
                blob.name.endswith(f"configuration_{VALUES_FILENAME}"),
                blob.name.endswith(f"configuration_{MANIFEST_FILENAME}"),
                blob.name.endswith(f"input_{VALUES_FILENAME}"),
                blob.name.endswith(f"input_{MANIFEST_FILENAME}"),
                blob.name.endswith("questions.json"),
            )
        )

    GoogleCloudStorageClient().download_all_files(
        local_path=local_path,
        cloud_path=cloud_path,
        filter=filter,
        recursive=True,
    )

    # Update the manifests with the local paths of the datasets.
    if download_datasets:
        for manifest_type in ("configuration_manifest", "input_manifest"):
            manifest_path = os.path.join(local_path, manifest_type + ".json")

            if not os.path.exists(manifest_path):
                continue

            manifest = Manifest.from_file(manifest_path)

            manifest.update_dataset_paths(
                path_generator=lambda dataset: os.path.join(local_path, f"{manifest_type}_datasets", dataset.name)
            )

            manifest.to_file(manifest_path)

    logger.info("Downloaded crash diagnostics from %r to %r.", cloud_path, local_path)


@octue_cli.group()
def deploy():
    """Deploy a python app to the cloud as an Octue service or digital twin."""


@deploy.command()
@click.option(
    "-c",
    "--service-config",
    type=click.Path(exists=True, dir_okay=False),
    default="octue.yaml",
    show_default=True,
    help="The path to an `octue.yaml` file defining the service to deploy.",
)
@click.option("--no-cache", is_flag=True, help="If provided, don't use the Docker cache when building the image.")
@click.option("--update", is_flag=True, help="If provided, allow updates to an existing service.")
@click.option(
    "--dataflow-job-only",
    is_flag=True,
    help="If provided, skip creating and running the build trigger and just deploy a pre-built image to Dataflow",
)
@click.option("--image-uri", type=str, default=None, help="The actual image URI to use when creating the Dataflow job.")
@click.option(
    "--revision-tag",
    type=str,
    default=None,
    help="A tag to use for this revision of the service (e.g. 1.3.7). This overrides the `OCTUE_SERVICE_REVISION_TAG` "
    "environment variable if it's present. If this option isn't given and the environment variable isn't present, a "
    "random 'cool name' tag is generated e.g 'curious-capybara'.",
)
def dataflow(service_config, no_cache, update, dataflow_job_only, image_uri, revision_tag):
    """Deploy a python app to Google Dataflow as an Octue service or digital twin."""
    if bool(importlib.util.find_spec("apache_beam")):
        # Import the Dataflow deployer only if the `apache-beam` package is available (due to installing `octue` with
        # the `dataflow` extras option).
        from octue.cloud.deployment.google.dataflow.deployer import DataflowDeployer
    else:
        raise ImportWarning(
            "To use this CLI command, you must install `octue` with the `dataflow` option e.g. "
            "`pip install octue[dataflow]`"
        )

    deployer = DataflowDeployer(service_config, revision_tag=revision_tag)

    if dataflow_job_only:
        deployer.create_streaming_dataflow_job(image_uri=image_uri, update=update)
        return

    deployer.deploy(no_cache=no_cache, update=update)


@deploy.command()
@click.argument("project_name")
@click.argument("service_namespace")
@click.argument("service_name")
@click.argument("push_endpoint")
@click.option(
    "--expiration-time",
    is_flag=False,
    default=None,
    show_default=True,
    help="The number of seconds of inactivity after which the subscription should expire. If not provided, no "
    "expiration time is applied to the subscription.",
)
@click.option(
    "--revision-tag",
    is_flag=False,
    default=None,
    show_default=True,
    help="The service revision tag (e.g. 1.0.7). If this option isn't given, a random 'cool name' tag is generated e.g"
    ". 'curious-capybara'.",
)
def create_push_subscription(
    project_name,
    service_namespace,
    service_name,
    push_endpoint,
    expiration_time,
    revision_tag,
):
    """Create a push subscription on Google Pub/Sub from the Octue service to the push endpoint. If a corresponding
    topic doesn't exist, it will be created. The subscription name is printed on completion.

    PROJECT_NAME is the name of the Google Cloud project in which the subscription will be created

    SERVICE_NAMESPACE is the namespace the service belongs to in kebab case

    SERVICE_NAME is the name of the service in kebab case, unique within its namespace

    PUSH_ENDPOINT is the HTTP/HTTPS endpoint of the service to push to. It should be fully formed and include the
    'https://' prefix
    """
    service_sruid = create_sruid(namespace=service_namespace, name=service_name, revision_tag=revision_tag)
    pub_sub_sruid = convert_service_id_to_pub_sub_form(service_sruid)

    topic = Topic(name=pub_sub_sruid, project_name=project_name)
    topic.create(allow_existing=True)

    if expiration_time:
        expiration_time = float(expiration_time)
    else:
        # Convert empty string to `None`.
        expiration_time = None

    subscription = Subscription(
        name=pub_sub_sruid,
        topic=topic,
        project_name=project_name,
        expiration_time=expiration_time,
        push_endpoint=push_endpoint,
    )

    subscription.create()
    click.echo(service_sruid)


def _add_monitor_message_to_file(path, monitor_message):
    """Add a monitor message to the file at the given path.

    :param str path: the path of the file to add the monitor message to
    :param dict monitor_message: the monitor message to add to the file
    :return None:
    """
    previous_messages = []

    if os.path.exists(path):
        try:
            with open(path) as f:
                previous_messages = json.load(f)
        except json.decoder.JSONDecodeError:
            pass

    previous_messages.append(monitor_message)

    with open(path, "w") as f:
        json.dump(previous_messages, f)


if __name__ == "__main__":
    args = sys.argv[1:] if len(sys.argv) > 1 else []
    octue_cli(args)



import base64
import collections.abc
import datetime

from google_crc32c import Checksum


EMPTY_STRING_HASH_VALUE = "AAAAAA=="

_HASH_PREPARATION_FUNCTIONS = {
    str: lambda attribute: attribute,
    int: str,
    float: str,
    type(None): lambda attribute: "None",
    datetime.datetime: str,
}


class Hashable:

    _ATTRIBUTES_TO_HASH = None
    _HASH_TYPE = "CRC32C"

    def __init__(self, immutable_hash_value=None, *args, **kwargs):
        self._immutable_hash_value = immutable_hash_value
        self._ATTRIBUTES_TO_HASH = self._ATTRIBUTES_TO_HASH or []
        super().__init__(*args, **kwargs)

    @classmethod
    def hash_non_class_object(cls, object_):
        """Use the Hashable class to hash an arbitrary object that isn't an attribute of a class instance.

        :param any object_:
        :return str:
        """

        class Holder(cls):
            _ATTRIBUTES_TO_HASH = ("object_",)

        holder = Holder()
        holder.object_ = object_
        return holder.hash_value

    @property
    def hash_value(self):
        """Get the hash of the instance.

        :return str:
        """
        if self._immutable_hash_value is None:
            return self._calculate_hash()

        return self._immutable_hash_value

    @hash_value.setter
    def hash_value(self, value):
        """Set the hash of the instance to a custom value.

        :param str value:
        :return None:
        """
        if self._immutable_hash_value is not None:
            raise ValueError(f"The hash of {self!r} is immutable - hash_value cannot be set.")

        self._immutable_hash_value = value

    def reset_hash(self):
        """Reset the hash value to the calculated hash (rather than whatever value has been set).

        :return None:
        """
        self._immutable_hash_value = None

    def _calculate_hash(self, hash_=None):
        """Calculate the hash of the sorted attributes in self._ATTRIBUTES_TO_HASH. If `hash_` is not `None` and is
        a hasher object, its in-progress hash will be updated, rather than starting from scratch.

        :param google_crc32c.Checksum hash_:
        :return str:
        """
        hash_ = hash_ or Checksum()

        for attribute_name in sorted(self._ATTRIBUTES_TO_HASH):
            attribute = getattr(self, attribute_name)

            try:
                items_to_hash = _HASH_PREPARATION_FUNCTIONS[type(attribute)](attribute)

            except KeyError:
                if isinstance(attribute, dict):
                    items_to_hash = self._prepare_dictionary_for_hashing(attribute_name, attribute)
                elif isinstance(attribute, collections.abc.Iterable):
                    items_to_hash = self._prepare_iterable_for_hashing(attribute_name, attribute)
                else:
                    raise TypeError(f"Attribute <{attribute_name!r}: {attribute!r}> cannot be hashed.")

            hash_.update(items_to_hash.encode())

        return base64.b64encode(hash_.digest()).decode("utf-8")

    def _prepare_iterable_for_hashing(self, attribute_name, attribute):
        """Prepare an iterable attribute for hashing, using the items' own `hash_value` attributes if available.

        :param str attribute_name:
        :param iter attribute:
        :return str:
        """
        items = tuple(attribute)

        if any(hasattr(item, "hash_value") for item in items):
            if not all(hasattr(item, "hash_value") for item in items):
                raise ValueError(f"Mixed types in attribute <{attribute_name!r}: {attribute!r}>")

            return self._sort_items_and_convert_to_string((item.hash_value for item in items), attribute_name)

        return self._sort_items_and_convert_to_string(items, attribute_name)

    def _prepare_dictionary_for_hashing(self, attribute_name, attribute):
        """Prepare a dictionary attribute for hashing, using its values' own `hash_value` attributes if available.

        :param str attribute_name:
        :param dict attribute:
        :return str:
        """
        items = tuple(attribute.items())

        if any(hasattr(item[1], "hash_value") for item in items):
            if not all(hasattr(item[1], "hash_value") for item in items):
                raise ValueError(f"Mixed types in attribute <{attribute_name!r}: {attribute!r}>")

            return self._sort_items_and_convert_to_string(
                ((key, value.hash_value) for key, value in items),
                attribute_name,
            )

        return self._sort_items_and_convert_to_string(items, attribute_name)

    @staticmethod
    def _sort_items_and_convert_to_string(items, attribute_name):
        """Sort the given items and cast them to a string.

        :param iter items:
        :param str attribute_name:
        :return str:
        """
        try:
            return str(sorted(items))

        except TypeError:
            raise TypeError(
                f"Attribute <{attribute_name!r}: {items!r}> needs to be sorted for consistent hash output, but this "
                f"failed."
            )



import copy
import json
import logging
from unittest.mock import patch

from octue.cloud import EXCEPTIONS_MAPPING
from octue.cloud.emulators._pub_sub import MockService, MockSubscriber, MockSubscription, MockTopic
from octue.resources import Analysis, Manifest, service_backends
from octue.utils.patches import MultiPatcher


logger = logging.getLogger(__name__)


class ChildEmulator:
    """An emulator for the `octue.resources.child.Child` class that sends the given messages to the parent for handling
    without contacting the real child or using Pub/Sub. Any messages a real child could produce are supported. `Child`
    instances can be replaced/mocked like-for-like by `ChildEmulator` without the parent knowing.

    :param str|None id: the ID of the child; a UUID is generated if none is provided
    :param dict|None backend: a dictionary including the key "name" with a value of the name of the type of backend (e.g. "GCPPubSubBackend") and key-value pairs for any other parameters the chosen backend expects; a mock backend is used if none is provided
    :param str internal_service_name: the name to give to the internal service used to ask questions to the child
    :param list(dict)|None messages: the list of messages to send to the parent
    :return None:
    """

    def __init__(self, id=None, backend=None, internal_service_name="local/local:local", messages=None):
        self.messages = messages or []

        backend = copy.deepcopy(backend or {"name": "GCPPubSubBackend", "project_name": "emulated-project"})
        backend_type_name = backend.pop("name")
        backend = service_backends.get_backend(backend_type_name)(**backend)

        self._child = MockService(service_id=id, backend=backend, run_function=self._emulate_analysis)
        self.id = self._child.id

        self._parent = MockService(
            backend=backend,
            service_id=internal_service_name,
            children={self._child.id: self._child},
        )

        self._message_handlers = {
            "delivery_acknowledgement": self._handle_delivery_acknowledgement,
            "heartbeat": self._handle_heartbeat,
            "log_record": self._handle_log_record,
            "monitor_message": self._handle_monitor_message,
            "exception": self._handle_exception,
            "result": self._handle_result,
        }

        self._valid_message_types = set(self._message_handlers.keys())

    @classmethod
    def from_file(cls, path):
        """Instantiate a child emulator from a JSON file at the given path. All/any/none of the instantiation arguments
        can be given in the file.

        :param str path: the path to a JSON file representing a child emulator
        :return ChildEmulator:
        """
        with open(path) as f:
            serialised_child_emulator = json.load(f)

        return cls(
            id=serialised_child_emulator.get("id"),
            backend=serialised_child_emulator.get("backend"),
            internal_service_name=serialised_child_emulator.get("internal_service_name"),
            messages=serialised_child_emulator.get("messages"),
        )

    def __repr__(self):
        """Represent a child emulator as a string.

        :return str:
        """
        return f"<{type(self).__name__}({self.id!r})>"

    @property
    def received_messages(self):
        """Get the messages received from the child.

        :return list(dict):
        """
        return self._parent.received_messages

    def ask(
        self,
        input_values=None,
        input_manifest=None,
        subscribe_to_logs=True,
        allow_local_files=False,
        handle_monitor_message=None,
        record_messages=True,
        question_uuid=None,
        timeout=86400,
    ):
        """Ask the child emulator a question and receive its emulated response messages. Unlike a real child, the input
         values and manifest are not validated against the schema in the child's twine as it is only available to the
         real child. Hence, the input values and manifest do not affect the messages returned by the emulator.

        :param any|None input_values: any input values for the question
        :param octue.resources.manifest.Manifest|None input_manifest: an input manifest of any datasets needed for the question
        :param bool subscribe_to_logs: if `True`, subscribe to logs from the child and handle them with the local log handlers
        :param bool allow_local_files: if `True`, allow the input manifest to contain references to local files - this should only be set to `True` if the child will have access to these local files
        :param callable|None handle_monitor_message: a function to handle monitor messages (e.g. send them to an endpoint for plotting or displaying) - this function should take a single JSON-compatible python primitive as an argument (note that this could be an array or object)
        :param bool record_messages: if `True`, record messages received from the child in the `received_messages` property
        :param str|None question_uuid: the UUID to use for the question if a specific one is needed; a UUID is generated if not
        :param float timeout: time in seconds to wait for an answer before raising a timeout error
        :raise TimeoutError: if the timeout is exceeded while waiting for an answer
        :return dict: a dictionary containing the keys "output_values" and "output_manifest"
        """
        with ServicePatcher():
            self._child.serve(allow_existing=True)

            subscription, _ = self._parent.ask(
                service_id=self._child.id,
                input_values=input_values,
                input_manifest=input_manifest,
                subscribe_to_logs=subscribe_to_logs,
                allow_local_files=allow_local_files,
                question_uuid=question_uuid,
            )

            return self._parent.wait_for_answer(
                subscription,
                handle_monitor_message=handle_monitor_message,
                record_messages=record_messages,
                service_name=self.id,
                timeout=timeout,
            )

    def _emulate_analysis(
        self,
        analysis_id,
        input_values,
        input_manifest,
        children,
        analysis_log_handler,
        handle_monitor_message,
        allow_save_diagnostics_data_on_crash,
    ):
        """Emulate analysis of a question by handling the messages given at instantiation in the order given.

        :param str|None analysis_id: UUID of analysis
        :param str|dict|None input_values: any input values for the question
        :param str|dict|octue.resources.manifest.Manifest|None input_manifest: an input manifest of any datasets needed for the question
        :param list(dict)|None children: a list of children for the child to use instead of its default children (if it uses children). These should be in the same format as in an app's app configuration file and have the same keys. (this is ignored by the emulator)
        :param logging.Handler|None analysis_log_handler: the `logging.Handler` instance which will be used to handle logs for this analysis run (this is ignored by the emulator)
        :param callable|None handle_monitor_message: a function that sends monitor messages to the parent that requested the analysis
        :param bool allow_save_diagnostics_data_on_crash: if `True`, allow the input values and manifest (and its datasets) to be saved if the analysis fails
        :return octue.resources.analysis.Analysis:
        """
        for message in self.messages:
            self._validate_message(message)
            handler = self._message_handlers[message["type"]]

            result = handler(
                message,
                analysis_id=analysis_id,
                input_values=input_values,
                input_manifest=input_manifest,
                children=children,
                analysis_log_handler=analysis_log_handler,
                handle_monitor_message=handle_monitor_message,
                allow_save_diagnostics_data_on_crash=allow_save_diagnostics_data_on_crash,
            )

            if result:
                return result

        # If no result message is included in the given messages, return an empty analysis.
        return Analysis(
            id=analysis_id,
            twine={},
            handle_monitor_message=handle_monitor_message,
            input_values=input_values,
            input_manifest=input_manifest,
            output_values=None,
            output_manifest=None,
        )

    def _validate_message(self, message):
        """Validate the given message to ensure it can be handled.

        :param dict message:
        :raise TypeError: if the message isn't a dictionary
        :raise ValueError: if the message doesn't contain a 'type' key or if the 'type' key maps to an invalid value
        :return None:
        """
        if not isinstance(message, dict):
            raise TypeError("Each message must be a dictionary.")

        if "type" not in message:
            raise ValueError(
                f"Each message must contain a 'type' key mapping to one of: {self._valid_message_types!r}."
            )

        if message["type"] not in self._valid_message_types:
            raise ValueError(
                f"{message['type']!r} is an invalid message type for the ChildEmulator. The valid types are: "
                f"{self._valid_message_types!r}."
            )

    def _handle_delivery_acknowledgement(self, message, **kwargs):
        """A no-operation handler for delivery acknowledgement messages (these messages are ignored by the child
        emulator).

        :param dict message: a dictionary containing the key "delivery_time"
        :param kwargs: this should be empty
        :return None:
        """
        logger.warning("Delivery acknowledgement messages are ignored by the ChildEmulator.")

    def _handle_heartbeat(self, message, **kwargs):
        """A no-operation handler for heartbeat messages (these messages are ignored by the child emulator).

        :param dict message: a dictionary containing the key "time"
        :param kwargs: this should be empty
        :return None:
        """
        logger.warning("Heartbeat messages are ignored by the ChildEmulator.")

    def _handle_log_record(self, message, **kwargs):
        """Convert the given message into a log record and pass it to the log handler.

        :param dict message: a dictionary containing a "log_record" key whose value is a dictionary representing a log record
        :param kwargs: this should be empty
        :raise TypeError: if the message can't be converted to a log record
        :return None:
        """
        try:
            log_record = message["log_record"]
        except KeyError:
            raise ValueError("Log record messages must include a 'log_record' key.")

        try:
            log_record["levelno"] = log_record.get("levelno", 20)
            log_record["levelname"] = log_record.get("levelname", "INFO")
            log_record["name"] = log_record.get("name", f"{__name__}.{type(self).__name__}")
            logger.handle(logging.makeLogRecord(log_record))

        except Exception:
            raise TypeError(
                "The 'log_record' key in a log record message must map to a dictionary that can be converted by "
                "`logging.makeLogRecord` to a `logging.LogRecord` instance."
            )

    def _handle_monitor_message(self, message, **kwargs):
        """Handle a monitor message with the given handler.

        :param dict message: a dictionary containing a "data" key mapped to a JSON-encoded string representing a monitor message. This monitor message will be handled by the monitor message handler
        :param kwargs: must include the "handle_monitor_message" key
        :return None:
        """
        kwargs.get("handle_monitor_message")(json.loads(message["data"]))

    def _handle_exception(self, message, **kwargs):
        """Raise the given exception.

        :param dict message: a dictionary representing the exception to be raised; it must include the "exception_type" and "exception_message" keys
        :param kwargs: this should be empty
        :raise ValueError: if the given exception cannot be raised
        :return None:
        """
        if "exception_type" not in message or "exception_message" not in message:
            raise ValueError(
                "The exception must be given as a dictionary containing the keys 'exception_type' and "
                "'exception_message'."
            )

        try:
            exception_type = EXCEPTIONS_MAPPING[message["exception_type"]]

        # Allow unknown exception types to still be raised.
        except KeyError:
            exception_type = type(message["exception_type"], (Exception,), {})

        raise exception_type(message["exception_message"])

    def _handle_result(self, message, **kwargs):
        """Return the result as an `Analysis` instance.

        :param dict message: a dictionary containing an "output_values" key and an "output_manifest" key
        :param kwargs: must contain the keys "analysis_id", "handle_monitor_message", "input_values", and "input_manifest"
        :raise ValueError: if the result doesn't contain the "output_values" and "output_manifest" keys
        :return octue.resources.analysis.Analysis: an `Analysis` instance containing the emulated outputs
        """
        output_manifest = message.get("output_manifest")

        if output_manifest and not isinstance(output_manifest, Manifest):
            output_manifest = Manifest.deserialise(output_manifest)

        try:
            return Analysis(
                id=kwargs["analysis_id"],
                twine={},
                handle_monitor_message=kwargs["handle_monitor_message"],
                input_values=kwargs["input_values"],
                input_manifest=kwargs["input_manifest"],
                output_values=message["output_values"],
                output_manifest=output_manifest,
            )

        except KeyError:
            raise ValueError(
                "Result messages must be dictionaries containing the keys 'output_values' and 'output_manifest'."
            )


class ServicePatcher(MultiPatcher):
    """A multi-patcher that provides the patches needed to run mock services.

    :return None:
    """

    def __init__(self):
        super().__init__(
            patches=[
                patch("octue.cloud.pub_sub.service.Topic", new=MockTopic),
                patch("octue.cloud.pub_sub.service.Subscription", new=MockSubscription),
                patch("octue.cloud.pub_sub.message_handler.SubscriberClient", new=MockSubscriber),
                patch("google.cloud.pubsub_v1.SubscriberClient", new=MockSubscriber),
            ]
        )



import json
import os

from octue.cloud.emulators.child import ChildEmulator
from octue.definitions import MANIFEST_FILENAME, VALUES_FILENAME
from octue.resources import Manifest


def load_test_fixture_from_crash_diagnostics(path):
    """Load a test fixture from service crash diagnostics downloaded using the `octue get-crash-diagnostics` CLI
    command. The configuration values, configuration manifest, input values, and input manifest are returned if
    available. A tuple of child emulators is returned if the service has children and asked questions to any of them.
    Each child emulator corresponds to one question asked to a child. The child emulators are in the same order as the
    questions were asked by the service during its analysis.

    :param str path: the path to a local directory containing downloaded crash diagnostics data
    :return (any|None, octue.resources.manifest.Manifest|None, any|None, octue.resources.manifest.Manifest|None, tuple(octue.cloud.emulators.child.ChildEmulator)): the configuration values, configuration manifest, input values, input manifest (each if available), and a tuple of child emulators corresponding to the questions asked by the service
    """
    return (
        _load_values(path, "configuration"),
        _load_manifest(path, "configuration"),
        _load_values(path, "input"),
        _load_manifest(path, "input"),
        _load_child_emulators(path),
    )


def _load_values(path, stage):
    """Load values from a JSON file in the directory specified by `path`.

    :param str path: the path to the crash diagnostics directory
    :param str stage: one of "configuration" or "input"
    :return any|None: the values or `None` if no relevant file exists in the directory specified by `path`
    """
    path = os.path.join(path, f"{stage}_{VALUES_FILENAME}")

    if not os.path.exists(path):
        return

    with open(path) as f:
        return json.load(f)


def _load_manifest(path, stage):
    """Load a manifest from a JSON file in the directory specified by `path`.

    :param str path: the path to the crash diagnostics directory
    :param str stage: one of "configuration" or "input"
    :return any|None: the manifest or `None` if no relevant file exists in the directory specified by `path`
    """
    path = os.path.join(path, f"{stage}_{MANIFEST_FILENAME}")

    if not os.path.exists(path):
        return

    return Manifest.from_file(path)


def _load_child_emulators(path):
    """Load child emulators from a JSON file in the directory specified by `path`.

    :param str path: the path to the crash diagnostics directory
    :return tuple(octue.resources.child.Child):
    """
    with open(os.path.join(path, "questions.json")) as f:
        questions = json.load(f)

    return tuple(ChildEmulator(id=question["id"], messages=question["messages"]) for question in questions)



import time
import warnings
from threading import Timer

from octue.utils.processes import run_logged_subprocess as moved_run_logged_sub_process


class RepeatingTimer(Timer):
    """A repeating version of the `threading.Timer` class."""

    def run(self):
        while not self.finished.is_set():
            time.sleep(self.interval)
            self.function(*self.args, **self.kwargs)


def run_logged_subprocess(command, logger, log_level="info", *args, **kwargs):
    """The deprecated version of `octue.utils.processes.run_logged_subprocess`.

    :param iter(str) command: command to run
    :param logging.Logger logger: logger to use to log stdout and stderr
    :param str log_level: level to log output at
    :raise CalledProcessError: if the subprocess fails (i.e. if it doesn't exit with a 0 return code)
    :return subprocess.CompletedProcess:
    """
    warnings.warn(
        DeprecationWarning(
            "`run_logged_subprocess` has been moved to `octue.utils.processes`. Importing it from "
            "`octue.utils.threads` is deprecated and the ability to do so will be removed soon."
        )
    )

    moved_run_logged_sub_process(command, logger, log_level=log_level, *args, **kwargs)



import json
import logging
import os


logger = logging.getLogger(__name__)


try:
    with open(os.path.join(os.path.dirname(__file__), "metadata", "version_compatibilities.json")) as f:
        VERSION_COMPATIBILITIES = json.load(f)
except FileNotFoundError:
    logger.warning("Version compatibility data could not be loaded.")
    VERSION_COMPATIBILITIES = {}


def is_compatible(parent_sdk_version, child_sdk_version):
    """Check if two versions of `octue` are compatible according to empirical testing. If there's no information on one
    of the versions, a warning is issued and they're treated as compatible.

    :param str parent_sdk_version: the semantic version of Octue SDK running the parent
    :param str child_sdk_version: the semantic version of Octue SDK running the child
    :return bool:
    """
    if (
        parent_sdk_version not in VERSION_COMPATIBILITIES
        or child_sdk_version not in VERSION_COMPATIBILITIES[parent_sdk_version]
    ):
        logger.warning(
            "No data on compatibility of parent SDK version %s and child SDK version %s.",
            parent_sdk_version,
            child_sdk_version,
        )
        return True

    return VERSION_COMPATIBILITIES[parent_sdk_version][child_sdk_version]


def warn_if_incompatible(parent_sdk_version, child_sdk_version):
    """Log a warning if the parent SDK version isn't compatible with the child SDK version, or if compatibility can't be
    checked due to an absence of version information for one of them.

    :param str|None parent_sdk_version: the version of the Octue SDK running locally / on the local service
    :param str|None child_sdk_version: the version of the Octue SDK running on the remote service
    :return None:
    """
    if not parent_sdk_version:
        missing_service_version_information = "parent"
    elif not child_sdk_version:
        missing_service_version_information = "child"
    else:
        missing_service_version_information = None

    if missing_service_version_information:
        logger.warning(
            "The %s couldn't be checked for compatibility with this service because its Octue SDK version wasn't "
            "provided. Please update it to the latest Octue SDK version.",
            missing_service_version_information,
        )
        return

    if not is_compatible(parent_sdk_version, child_sdk_version):
        logger.warning(
            "The parent's Octue SDK version %s is incompatible with the child's version %s. Please update either or "
            "both to the latest version.",
            parent_sdk_version,
            child_sdk_version,
        )



import re
from datetime import datetime, timezone

from stringcase import snakecase


def clean(data, date):
    """Clean up data from meteorological mast anemometers and wind vanes...

    ... or rename the function and do pretty much whatever you want with the data!

    In this ultra-simple example, we'll parse the time and add the date to it (timestamps are much more useful when
    complete!) as well as correcting a typo in column headings

    :param pandas.DataFrame data: Pandas dataframe containing imported, uncleaned data
    :return pandas.DataFrame: The same dataframe, cleaned
    """
    # Add a proper datestamp column (combining the time from the time column with the date from the metadata).
    cleaned_timestamps = []
    for ts in data["TimeStamp"]:
        # Decimal minutes. REALLY? This is why our work is so hard. Learn from the gross lack of foresight of whoever
        # made this meteorological mast, and put data into sensible formats (ISO formatted timestamp would've been
        # really helpful here!).
        hms = [int(value) for value in re.split(r"[:.]", ts)]
        hms[2] = hms[2] * 6
        cleaned_timestamps.append(
            datetime(
                year=date.year,
                month=date.month,
                day=date.day,
                hour=hms[0],
                minute=hms[0],
                second=hms[1],
                tzinfo=timezone.utc,  # UTC is a best guess since no timezone info supplied with the data.
            )
        )

    # Replace the near-useless decimal-minutes-no-date-no-timezone with a cleaned, complete datestamp.
    data.pop("TimeStamp")
    data["time_stamp"] = cleaned_timestamps  # If storage space is a premium, convert this to posix time

    # And clean up typos made by an overly stressed out junior engineer who is too busy to care about the pain this
    # has caused you finding it.
    data["Barometer_1"] = data.pop("Barmoeter_1")

    # And convert column names to snake_case, so they make sensible variable names without spaces etc.
    data.rename(columns=lambda old_name: snakecase(old_name))

    return data
